package djmil.cordacheckers.contracts;

import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleInputState;
import static djmil.cordacheckers.contracts.UtxoLedgerTransactionUtil.getSingleOutputState;

import java.util.List;

import djmil.cordacheckers.checkers.Move;
import djmil.cordacheckers.states.GameBoardState;
import djmil.cordacheckers.states.GameProposalState;
import djmil.cordacheckers.states.GameResultState;
import net.corda.v5.base.annotations.ConstructorForDeserialization;
import net.corda.v5.base.annotations.CordaSerializable;
import net.corda.v5.ledger.utxo.Command;
import net.corda.v5.ledger.utxo.transaction.UtxoLedgerTransaction;

public class GameCommand implements Command {
    @CordaSerializable
    public static enum Action {
        GAME_PROPOSAL_CREATE,
        GAME_PROPOSAL_CANCEL,
        GAME_PROPOSAL_ACCEPT,
        GAME_PROPOSAL_REJECT,

        MOVE,
        DRAW_REQUEST,
        DRAW_ACCEPT,
        DRAW_REJECT,
        SURRENDER,
        CLAIM_VICTORY;
    }

    private final Action action;
    private final List<Integer> move;

    public static class ActionException extends RuntimeException {
        public ActionException() {
            super("Bad GameCommand.Action value");
        }
    }

    // Serialisation service requires a default constructor
    public GameCommand() {
        this.action = null;
        this.move = null;
    }

    public GameCommand(Action action) {
        if (action == Action.MOVE)
            throw new ActionException();

        this.action = action;
        this.move = null;
    }

    public GameCommand(List<Integer> move) {
        this.action = Action.MOVE;
        this.move = move;
    }

    @ConstructorForDeserialization
    public GameCommand(Action action, List<Integer> move) {
        this.action = action;
        this.move = move;
    }

    public Action getAction() {
        return action;
    }

    public List<Integer> getMove() {
        return move;
    }

    public void validateGameProposalCreate(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().isEmpty(), CREATE_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, CREATE_OUTPUT_STATE);

        final GameProposalState gameProposal = getSingleOutputState(trx, GameProposalState.class);

        /*
         * Major command logic check 
         */
        requireThat(gameProposal.getIssuerName().compareTo(gameProposal.getBlackPlayer()) == 0 ||
                    gameProposal.getIssuerName().compareTo(gameProposal.getWhitePlayer()) == 0,
            "GameProposal.Issuer must be either Black or White player");
    }

    public void validateGameProposalAccept(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size()  == 1, ACCEPT_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, ACCEPT_OUTPUT_STATE);

        final GameProposalState inGameProposal = getSingleInputState (trx, GameProposalState.class);
        final GameBoardState outGameBoard      = getSingleOutputState(trx, GameBoardState.class);

        requireThat(outGameBoard.getWhitePlayer().compareTo(inGameProposal.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getBlackPlayer().compareTo(inGameProposal.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getParticipants().containsAll(inGameProposal.getParticipants()), IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getGameUuid().compareTo(inGameProposal.getGameUuid()) == 0, UUID_MUST_NOT_CHANGE);

        /*
         * Major command logic check 
         */
        requireThat(outGameBoard.getBoard().size() > 0, "GameBoard initial state was not found");
    }

    public void validateGameProposalReject(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, REJECT_INPUT_STATE);
        requireThat(trx.getOutputContractStates().isEmpty(), REJECT_OUTPUT_STATE);

        getSingleInputState(trx, GameProposalState.class);
    }

    public void validateGameProposalCancel(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size() == 1, CANCEL_INPUT_STATE);
        requireThat(trx.getOutputContractStates().isEmpty(), CANCEL_OUTPUT_STATE);
        
        getSingleInputState(trx, GameProposalState.class);
    }

    public void validateMove(UtxoLedgerTransaction trx, List<Integer> move) {
        requireThat(trx.getInputContractStates().size()  == 1, MOVE_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, MOVE_OUTPUT_STATE);

        final var inGameBoard  = getSingleInputState (trx, GameBoardState.class);
        final var outGameBoard = getSingleOutputState(trx, GameBoardState.class);

        requireThat(outGameBoard.getWhitePlayer().compareTo(inGameBoard.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getBlackPlayer().compareTo(inGameBoard.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getParticipants().containsAll(inGameBoard.getParticipants()), IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getGameUuid().compareTo(inGameBoard.getGameUuid()) == 0, UUID_MUST_NOT_CHANGE);

        /*
         * Major command logic check 
         */
        final var newGameBoard = new GameBoardState(inGameBoard, move, outGameBoard.getMessage());
        requireThat(outGameBoard.equals(newGameBoard), "Unexpected output state");
        requireThat(inGameBoard.isDrawRequested() == false, "Draw was requested");
        requireThat(outGameBoard.isDrawRequested() == false, "Draw can not be requested during move");
    }

    public void validateSurrender(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size()  == 1, SURRENDER_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, SURRENDER_OUTPUT_STATE);

        final var inGameBoard   = getSingleInputState (trx, GameBoardState.class);
        final var outGameResult = getSingleOutputState(trx, GameResultState.class);

        requireThat(outGameResult.getWhitePlayer().compareTo(inGameBoard.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getBlackPlayer().compareTo(inGameBoard.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getParticipants().containsAll(inGameBoard.getParticipants()), IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getGameUuid().compareTo(inGameBoard.getGameUuid()) == 0, UUID_MUST_NOT_CHANGE);

        /*
         * Major command logic check 
         */
        requireThat(outGameResult.getLooserName().compareTo(outGameResult.getBlackPlayer()) == 0 ||
                    outGameResult.getLooserName().compareTo(outGameResult.getWhitePlayer()) == 0,
            "Surenderer must be either Black or White player");

        requireThat(outGameResult.getTotalMoves() == inGameBoard.getMoveNumber(), BAD_TOTAL_MOVES);
    }

    public void validateClaimVictory(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size()  == 1, VICTORY_INPUT_STATE);
        requireThat(trx.getOutputContractStates().size() == 1, VICTORY_OUTPUT_STATE);

        final var inGameBoard   = getSingleInputState (trx, GameBoardState.class);
        final var outGameResult = getSingleOutputState(trx, GameResultState.class);

        requireThat(outGameResult.getWhitePlayer().compareTo(inGameBoard.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getBlackPlayer().compareTo(inGameBoard.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getParticipants().containsAll(inGameBoard.getParticipants()), IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getGameUuid().compareTo(inGameBoard.getGameUuid()) == 0, UUID_MUST_NOT_CHANGE);

        /*
         * Major command logic check 
         */
        final var possibleMoves = Move.getPossibleMoves(inGameBoard.getBoard(), inGameBoard.getActiveColor());
        requireThat(possibleMoves.isEmpty(), "Victory condition violation");

        requireThat(outGameResult.getWinnerName().compareTo(inGameBoard.getIdelPlayerName()) == 0, "Bad winner name");
        requireThat(outGameResult.getTotalMoves() == inGameBoard.getMoveNumber(), BAD_TOTAL_MOVES);
    }

    public void validateDrawRequest(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size()  == 1, Action.DRAW_REQUEST + BAD_IN_STATE_SIZE);
        requireThat(trx.getOutputContractStates().size() == 1, Action.DRAW_REQUEST + BAD_OUT_STATE_SIZE);

        final var inGameBoard  = getSingleInputState (trx, GameBoardState.class);
        final var outGameBoard = getSingleOutputState(trx, GameBoardState.class);

        requireThat(outGameBoard.getWhitePlayer().compareTo(inGameBoard.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getBlackPlayer().compareTo(inGameBoard.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getParticipants().containsAll(inGameBoard.getParticipants()), IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getGameUuid().compareTo(inGameBoard.getGameUuid()) == 0, UUID_MUST_NOT_CHANGE);

        /*
         * Major command logic check 
         */
        requireThat(isGameBoardEqual(inGameBoard, outGameBoard) == true, GAME_BOARD_STATE_MUST_NOT_CHANGE);
        requireThat(inGameBoard.getDrawRequest() == null, "Draw can be requested only once per move");
        requireThat(outGameBoard.isDrawRequested() == true, DRAW_REQ_EXPECTED);
    }

    public void validateDrawDecline(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size()  == 1, Action.DRAW_REJECT + BAD_IN_STATE_SIZE);
        requireThat(trx.getOutputContractStates().size() == 1, Action.DRAW_REJECT + BAD_OUT_STATE_SIZE);

        final var inGameBoard  = getSingleInputState (trx, GameBoardState.class);
        final var outGameBoard = getSingleOutputState(trx, GameBoardState.class);

        requireThat(outGameBoard.getWhitePlayer().compareTo(inGameBoard.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getBlackPlayer().compareTo(inGameBoard.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getParticipants().containsAll(inGameBoard.getParticipants()), IN_OUT_PARTICIPANTS);
        requireThat(outGameBoard.getGameUuid().compareTo(inGameBoard.getGameUuid()) == 0, UUID_MUST_NOT_CHANGE);

        /*
         * Major command logic check 
         */
        requireThat(isGameBoardEqual(inGameBoard, outGameBoard) == true, GAME_BOARD_STATE_MUST_NOT_CHANGE);
        requireThat(inGameBoard.isDrawRequested() == true, DRAW_REQ_EXPECTED);
        requireThat(outGameBoard.isDrawRequested() == false, "Draw decline expected");
    }

    public void validateDrawAcquire(UtxoLedgerTransaction trx) {
        requireThat(trx.getInputContractStates().size()  == 1, Action.DRAW_ACCEPT + BAD_IN_STATE_SIZE);
        requireThat(trx.getOutputContractStates().size() == 1, Action.DRAW_ACCEPT + BAD_OUT_STATE_SIZE);

        final var inGameBoard   = getSingleInputState (trx, GameBoardState.class);
        final var outGameResult = getSingleOutputState(trx, GameResultState.class);

        requireThat(outGameResult.getWhitePlayer().compareTo(inGameBoard.getWhitePlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getBlackPlayer().compareTo(inGameBoard.getBlackPlayer()) == 0, IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getParticipants().containsAll(inGameBoard.getParticipants()), IN_OUT_PARTICIPANTS);
        requireThat(outGameResult.getGameUuid().compareTo(inGameBoard.getGameUuid()) == 0, UUID_MUST_NOT_CHANGE);

        /*
         * Major command logic check 
         */
        requireThat(inGameBoard.isDrawRequested() == true, DRAW_REQ_EXPECTED);
        requireThat(outGameResult.getWinnerName() == null, "Draw can not have winner");
    }

    public static void requireThat(boolean asserted, String errorMessage) {
        if (!asserted) {
            throw new IllegalStateException("Failed requirement: " + errorMessage);
        }
    }

    private boolean isGameBoardEqual(GameBoardState a, GameBoardState b) {
        return a.getBoard().equals(b.getBoard()) &&
                a.getActiveColor() == b.getActiveColor() &&
                a.getMoveNumber() == b.getMoveNumber();
    }

    static final String REQUIRE_SINGLE_COMMAND = "Require a single command";
    static final String IN_OUT_PARTICIPANTS = "Output participants should include all input participants";

    static final String CREATE_INPUT_STATE  = "Create command should have no input states";
    static final String CREATE_OUTPUT_STATE = "Create command should output exactly one GameProposal state";

    static final String REJECT_INPUT_STATE  = "Reject command should have exactly one GameProposal input state";
    static final String REJECT_OUTPUT_STATE = "Reject command should have no output states";

    static final String CANCEL_INPUT_STATE  = "Cancel command should have exactly one GameProposal input state";
    static final String CANCEL_OUTPUT_STATE = "Cancel command should have no output states";

    static final String ACCEPT_INPUT_STATE  = "Accept command should have exactly one GameProposal input state";
    static final String ACCEPT_OUTPUT_STATE = "Accept command should have exactly one GameBoard output state";

    static final String SURRENDER_INPUT_STATE = "SURRENDER command should have exactly one GameBoardState input state";
    static final String SURRENDER_OUTPUT_STATE = "SURRENDER command should have exactly one GameResultState output state";

    static final String MOVE_INPUT_STATE = "MOVE command should have exactly one GameBoardState input state";
    static final String MOVE_OUTPUT_STATE = "MOVE command should have exactly one GameBoardState output state";
    static final String MOVE_OUT_BOARD = "MOVE command: move checkers rules violation";
    static final String MOVE_OUT_COLOR = "MOVE command: moveColor checkers rules violation";

    static final String VICTORY_INPUT_STATE  = "VICTORY command should have exactly one GameBoardState input state";
    static final String VICTORY_OUTPUT_STATE = "VICTORY command should have exactly one GameResultState output state";

    static final String BAD_TOTAL_MOVES = "Wrong number of total moves";
    
    static final String BAD_IN_STATE_SIZE  = " - wrong input state size";
    static final String BAD_OUT_STATE_SIZE = " - wrong output state size";

    static final String DRAW_REQ_EXPECTED = "Draw request expected";
    static final String GAME_BOARD_STATE_MUST_NOT_CHANGE = "GameBoard state must not be changed";
    static final String UUID_MUST_NOT_CHANGE = "UUID must not be changed";
}
